/** Utilities for Jupyterlab Bookmarks plugin.
 * Copyright (c) 2020, Dr. Nandor Poka, All rights reserved.
 */
import { ISettingRegistry } from '@jupyterlab/settingregistry';
import { INotebookTracker, NotebookPanel, INotebookModel } from '@jupyterlab/notebook';
import { DocumentRegistry } from '@jupyterlab/docregistry';
import { CommandRegistry } from '@lumino/commands';
import { Bookmark } from './bookmark';
import { ILauncher } from '@jupyterlab/launcher';
/**
 * Load the settings for this extension. Sets the `bookmarks` to the bookmarks stored in the settings.
 *
 * @param settings `ISettingsRegistry.Isettings` - Extension settings
 */
export declare function loadSetting(settings: ISettingRegistry.ISettings): void;
/**
 * Updates the JupyterLab apps `commands` variable with the command that is assigned to launch the bookmark.
 * @param commands - The `CommandRegistry` instance to add the command to.
 * @param bookmarkItem - `string[]` that has the information about the bookmark.
 * @returns `void`
 */
export declare function updateCommands(commands: CommandRegistry, bookmarkItem: Bookmark): void;
/**
 * Update the plugin's settings with the bookmark.
 * @async
 * @param bookmarkItem - `Bookmark` that stores the bookmark to be saved to the settings.
 * @returns `Promise<void>`
 */
export declare function updateSettings(bookmarkItem?: Bookmark): Promise<void>;
/**
 * Updates the `Launcher` with the bookmark.
 * @param launcher - The `Launcher` instance to update.
 * @param bookmarkItem - string[] that holds the bookmark to add.
 * @returns `void`
 */
export declare function updateLauncher(launcher: ILauncher, bookmarkItem: Bookmark): void;
/**
 * Updates the plugin's menu with the bookmark.
 * @param bookmarkItem - string[] that holds the bookmark to add.
 * @returns `void`
 */
export declare function updateMenu(bookmarkItem: Bookmark): void;
/**
 * Deletes the bookmark, and updates the `Launcher` the `Menu` and the `commands`.
 * @async
 * @param bookmarkToDelete - `string[]` the bookmark to delete.
 * @returns `Promise<void>
 */
export declare function deleteBookmark(bookmarkToDelete: string): Promise<void>;
/**
 * Adds new category. Duplicates cannot be added.
 * @param categoryToAdd `string` - the name of the category to add.
 * @param silent `boolean` - `true` if suppress warning for duplicate category.
 *
 */
export declare function addCategory(categoryToAdd: string, silent?: boolean): void;
/**
 * Deletes a category. Disposes launcher items in the category and moves bookmarks from
 * the category being deleted to Uncategorized`.
 * @param categoryToDelete `string` - the name of the category to delete.
 */
export declare function deleteCategory(categoryToDelete: string): void;
/** Adds a `bookmarkItem`, to the `commands` list, to the Launcher and to the Bookmarks menu.
 * @async
 * @param commands - A `CommandRegistry` instance that holds the commands registered in the JL app.
 * @param launcher - The `Launcher` instance to add the bookmark to.
 * @param bookmarkItem - `string[]`. The bookmarkItem to be added.
 * @param skipDuplicateCheck - `boolean`. Optional parameter to skip or not to skip
 * duplicate check.
 * @returns `Promise<boolean>`
 */
export declare function addBookmark(commands: CommandRegistry, launcher: ILauncher, bookmarkItem: Bookmark, skipDuplicateCheck?: boolean, startup?: boolean): Promise<boolean>;
/**
 * Adds a bookmark item from the context menu of an open notebook.
 * @async
 * @param commands - A `CommandRegistry` instance that holds the commands registered in the JL app.
 * @param launcher - The `Launcher` instance to add the bookmark to.
 * @param currentDocName - `string` the name of the open notebbook
 * @param currentDocPath - `string` the path of the open notebook
 * @param category - `string` the name of the category to add the bookmark to. Empty string if no category (`''`).
 * @returns `Promise<void>`
 */
export declare function addBookmarkItem(commands: CommandRegistry, launcher: ILauncher, currentDocName: string, currentDocPath: string, category: string): Promise<void>;
/**
 * Synchronises a notebook that has been opened from a temporary location back to it's original instance when the file is saved.
 * Makes a HTTP call to the plugins server extension that does the actual copying.
 * @param bookmarkedNotebookModel - `DocumentRegistry.IContext<INotebookModel>` the context object for the current NotebookWdiget.
 */
export declare function syncBookmark(bookmarkedNotebookModel: DocumentRegistry.IContext<INotebookModel>): void;
/**
 * Sets up the autosyncing for a notebook, by connecting the `syncBookmark` function to the notebook's context's fileChanged signal.
 * @param notebookPanel - `NotebookPanel`, the widget hosting the notebook.
 * @param notebookTracker - `INotebookTracker` instance as the sender
 */
export declare function addAutoSyncToBookmark(notebookTracker: INotebookTracker, notebookPanel: NotebookPanel): void;
/**
 * Compares two `Map<string, Bookmark>` instances if they are the same or not.
 * Returns `True` if and only if the two maps have the exact same entries.
 * For this the function compares all properties of the Bookmarks in the maps one by one.
 * Immediately returns `false` if the `Maps` are of different sizes, and
 * if the one of the Maps doesn't have an entry from the other.
 * @param persistentBookmarks
 * @param settingsBookmarks
 * @returns `boolean`
 */
export declare function compareBookmarkMaps(persistentBookmarks: Map<string, Bookmark>, settingsBookmarks: Map<string, Bookmark>): boolean;
/**
 * Imports the bookmarks stored in the JSON file stored in the argument.
 * @param bookmarkFile
 */
export declare function importBookmarks(bookmarkFile: File): void;
/**
 * Exports current bookmarks. Opens up browser native save dialog thus
 * the user can save the file in the desired location.
 */
export declare function exportBookmarks(): Promise<void>;
